/* automatically generated from ./regs.dat - do not edit */
static const long reg_flags[] = {
    0,
    REG8,
    REG_AL,
    REG_AX,
    REG8,
    REG8,
    REG16,
    REG16,
    REG8,
    REG_CL,
    REG_CREG,
    REG_CREG,
    REG_CREG,
    REG_CREG,
    REG_CREG,
    REG_CREG,
    REG_CREG,
    REG_CREG,
    REG_CS,
    REG_CX,
    REG8,
    REG16,
    REG_DL,
    REG_DREG,
    REG_DREG,
    REG_DREG,
    REG_DREG,
    REG_DREG,
    REG_DREG,
    REG_DREG,
    REG_DREG,
    REG_DESS,
    REG_DX,
    REG_EAX,
    REG32,
    REG32,
    REG_ECX,
    REG32,
    REG_EDX,
    REG_DESS,
    REG32,
    REG32,
    REG_FSGS,
    REG_FSGS,
    MMXREG,
    MMXREG,
    MMXREG,
    MMXREG,
    MMXREG,
    MMXREG,
    MMXREG,
    MMXREG,
    REG_SEG67,
    REG_SEG67,
    REG16,
    REG16,
    REG_DESS,
    FPU0,
    FPUREG,
    FPUREG,
    FPUREG,
    FPUREG,
    FPUREG,
    FPUREG,
    FPUREG,
    REG_TREG,
    REG_TREG,
    REG_TREG,
    REG_TREG,
    REG_TREG,
    REG_TREG,
    REG_TREG,
    REG_TREG,
    XMMREG,
    XMMREG,
    XMMREG,
    XMMREG,
    XMMREG,
    XMMREG,
    XMMREG,
    XMMREG
};
